from langchain.tools import BaseTool
from time import sleep
from rxn4chemistry import RXN4ChemistryWrapper
from chemcrow.utils import is_smiles

class RXNPredict(BaseTool):
    name = "RXNPredict"
    description = (
        "Predict the outcome of a chemical reaction. "
        "Takes as input the SMILES of the reactants separated by a dot '.', "
        "returns SMILES of the products."
    )
    rxn4chem_api_key: str = ''
    rxn4chem: RXN4ChemistryWrapper = None
    base_url: str = "https://rxn.res.ibm.com"
    sleep_time: str = 5

    def __init__(self, rxn4chem_api_key):
        super(RXNPredict, self).__init__()

        self.rxn4chem_api_key = rxn4chem_api_key
        self.rxn4chem = RXN4ChemistryWrapper(
            api_key=self.rxn4chem_api_key,
            base_url=self.base_url
        )
        self.rxn4chem.create_project("ChemCrow")

    def _run(self, reactants: str) -> str:

        # Check that input is smiles
        if not is_smiles(reactants):
            return "Incorrect input."

        # Send until valid response
        while True:
            sleep(self.sleep_time)
            response = self.rxn4chem.predict_reaction(reactants)
            if "prediction_id" in response.keys():
                break

        status = "WAITING"
        while True:
            sleep(self.sleep_time)
            results = self.rxn4chem.get_predict_reaction_results(
                response["prediction_id"]
            )
            if "payload" in results["response"].keys():
                break

        res_dict = results["response"]["payload"]["attempts"][0]

        rxn = res_dict["smiles"]
        product = res_dict["productMolecule"]["smiles"]

        return product



import pytest

from chemcrow.tools.databases import PatentCheck, Query2CAS, Query2SMILES
from chemcrow.utils import canonical_smiles


@pytest.fixture
def singlemol():
    # Single mol
    return "O=C1N(C)C(C2=C(N=CN2C)N1C)=O"


@pytest.fixture
def molset1():
    # Set of mols
    return "O=C1N(C)C(C2=C(N=CN2C)N1C)=O.CC(C)c1ccccc1"


@pytest.fixture
def molset2():
    # Set of mols
    return "O=C1N(C)C(C2=C(N=CN2C)N1C)=O.O=C1N(C)C(C2=C(N=CN2C)N1CCC)=O"


@pytest.fixture
def single_iupac():
    # Test with a molecule with iupac name
    return "4-(4-hydroxyphenyl)butan-2-one"


@pytest.fixture
def common_name():
    # Test with a molecule with common name
    return "caffeine"


@pytest.fixture
def choline():
    # Test with a molecule in clintox
    return "CCCCCCCCC[NH+]1C[C@@H]([C@H]([C@@H]([C@H]1CO)O)O)O"


# Query2SMILES


def test_q2s_iupac(single_iupac):
    tool = Query2SMILES()
    out = tool(single_iupac)
    assert out == "CC(=O)CCc1ccc(O)cc1"


def test_q2s_cafeine(singlemol):
    tool = Query2SMILES()
    out = tool("caffeine")
    assert out == canonical_smiles(singlemol)


def test_q2s_fail(molset1):
    tool = Query2SMILES()
    out = tool(molset1)
    assert out.endswith("input one molecule at a time.")


# Query2CAS


def test_q2cas_iupac(single_iupac):
    tool = Query2CAS()
    out = tool(single_iupac)
    assert out == "5471-51-2"


def test_q2cas_cafeine(singlemol):
    tool = Query2CAS()
    out = tool(singlemol)
    assert out == "58-08-2"


def test_q2cas_badinp():
    tool = Query2CAS()
    out = tool("nomol")
    assert out.endswith("no Pubchem entry")


# PatentCheck


def test_patentcheck(singlemol):
    tool = PatentCheck()
    patented = tool(singlemol)
    assert patented == "Patented"


def test_patentcheck_iupac(single_iupac):
    tool = PatentCheck()
    patented = tool(single_iupac)
    assert patented == "Invalid SMILES string"


def test_patentcheck_not(choline):
    tool = PatentCheck()
    patented = tool(choline)
    assert patented == "Novel"



import molbloom
import requests
from langchain.tools import BaseTool
from rdkit import Chem

from chemcrow.utils import *


class Query2SMILES(BaseTool):
    name = "Name2SMILES"
    description = "Input a molecule name, returns SMILES."
    url: str = None

    def __init__(
        self,
    ):
        super(Query2SMILES, self).__init__()
        self.url = "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/name/{}/{}"

    def _run(self, query: str) -> str:
        """This function queries the given molecule name and returns a SMILES string from the record"""
        """Useful to get the SMILES string of one molecule by searching the name of a molecule. Only query with one specific name."""

        # query the PubChem database
        r = requests.get(self.url.format(query, "property/IsomericSMILES/JSON"))
        # convert the response to a json object
        data = r.json()
        # return the SMILES string
        try:
            smi = data["PropertyTable"]["Properties"][0]["IsomericSMILES"]
        except KeyError:
            return "Could not find a molecule matching the text. One possible cause is that the input is incorrect, input one molecule at a time."
        # remove salts
        return Chem.CanonSmiles(largest_mol(smi))

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError()


class Query2CAS(BaseTool):
    name = "Mol2CAS"
    description = "Input molecule (name or SMILES), returns CAS number."
    url_cid: str = None
    url_data: str = None

    def __init__(
        self,
    ):
        super(Query2CAS, self).__init__()
        self.url_cid = (
            "https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/{}/{}/cids/JSON"
        )
        self.url_data = (
            "https://pubchem.ncbi.nlm.nih.gov/rest/pug_view/data/compound/{}/JSON"
        )

    def _run(self, query: str) -> str:
        try:
            mode = "name"
            if is_smiles(query):
                mode = "smiles"
            url_cid = self.url_cid.format(mode, query)
            cid = requests.get(url_cid).json()["IdentifierList"]["CID"][0]
            url_data = self.url_data.format(cid)
            data = requests.get(url_data).json()
        except (requests.exceptions.RequestException, KeyError):
            return "Invalid molecule input, no Pubchem entry"

        try:
            for section in data["Record"]["Section"]:
                if section.get("TOCHeading") == "Names and Identifiers":
                    for subsection in section["Section"]:
                        if subsection.get("TOCHeading") == "Other Identifiers":
                            for subsubsection in subsection["Section"]:
                                if subsubsection.get("TOCHeading") == "CAS":
                                    return subsubsection["Information"][0]["Value"][
                                        "StringWithMarkup"
                                    ][0]["String"]
        except KeyError:
            return "Invalid molecule input, no Pubchem entry"

        return "CAS number not found"

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError()


class PatentCheck(BaseTool):
    name = "PatentCheck"
    description = "Input SMILES, returns if molecule is patented"

    def _run(self, smiles: str) -> str:
        """Checks if compound is patented. Give this tool only one SMILES string"""
        try:
            r = molbloom.buy(smiles, canonicalize=True, catalog="surechembl")
        except:
            return "Invalid SMILES string"
        if r:
            return "Patented"
        else:
            return "Novel"

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError()



from langchain.tools import BaseTool
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem, rdMolDescriptors


class MolSimilarity(BaseTool):
    name = "MolSimilarity"
    description = (
        "Input two molecule SMILES (separated by '.'), returns Tanimoto similarity."
    )

    def __init__(self):
        super(MolSimilarity, self).__init__()

    def _run(self, smiles_pair: str) -> str:
        smi_list = smiles_pair.split(".")
        if len(smi_list) != 2:
            return "Input error, please input two smiles strings separated by '.'"
        else:
            smiles1, smiles2 = smi_list

        try:
            mol1 = Chem.MolFromSmiles(smiles1)
            mol2 = Chem.MolFromSmiles(smiles2)
            fp1 = AllChem.GetMorganFingerprintAsBitVect(mol1, 2, nBits=2048)
            fp2 = AllChem.GetMorganFingerprintAsBitVect(mol2, 2, nBits=2048)
            similarity = DataStructs.TanimotoSimilarity(fp1, fp2)

            sim_score = {
                0.9: "very similar",
                0.8: "similar",
                0.7: "somewhat similar",
                0.6: "not very similar",
                0: "not similar",
            }
            if similarity == 1:
                return "Error: Input Molecules Are Identical"
            else:
                val = sim_score[
                    max(key for key in sim_score.keys() if key <= round(similarity, 1))
                ]
                message = f"The Tanimoto similarity between {smiles1} and {smiles2} is {round(similarity, 4)},\
                indicating that the two molecules are {val}."
            return message
        except (TypeError, ValueError, AttributeError):
            return "Error: Not a valid SMILES string"

    async def _arun(self, smiles_pair: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError()


class SMILES2Weight(BaseTool):
    name = "SMILES2Weight"
    description = "Input SMILES, returns molecular weight."

    def __init__(
        self,
    ):
        super(SMILES2Weight, self).__init__()

    def _run(self, smiles: str) -> str:
        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            return "Invalid SMILES string"
        mol_weight = rdMolDescriptors.CalcExactMolWt(mol)
        return mol_weight

    async def _arun(self, smiles: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError()


class FuncGroups(BaseTool):
    name = "FunctionalGroups"
    description = "Input SMILES, return list of functional groups in the molecule."
    dict_fgs: dict = None

    def __init__(
        self,
    ):
        super(FuncGroups, self).__init__()

        # List obtained from https://github.com/rdkit/rdkit/blob/master/Data/FunctionalGroups.txt
        self.dict_fgs = {
            "furan": "o1cccc1",
            "aldehydes": " [CX3H1](=O)[#6]",
            "esters": " [#6][CX3](=O)[OX2H0][#6]",
            "ketones": " [#6][CX3](=O)[#6]",
            "amides": " C(=O)-N",
            "thiol groups": " [SH]",
            "alcohol groups": " [OH]",
            "methylamide": "*-[N;D2]-[C;D3](=O)-[C;D1;H3]",
            "carboxylic acids": "*-C(=O)[O;D1]",
            "carbonyl methylester": "*-C(=O)[O;D2]-[C;D1;H3]",
            "terminal aldehyde": "*-C(=O)-[C;D1]",
            "amide": "*-C(=O)-[N;D1]",
            "carbonyl methyl": "*-C(=O)-[C;D1;H3]",
            "isocyanate": "*-[N;D2]=[C;D2]=[O;D1]",
            "isothiocyanate": "*-[N;D2]=[C;D2]=[S;D1]",
            "nitro": "*-[N;D3](=[O;D1])[O;D1]",
            "nitroso": "*-[N;R0]=[O;D1]",
            "oximes": "*=[N;R0]-[O;D1]",
            "Imines": "*-[N;R0]=[C;D1;H2]",
            "terminal azo": "*-[N;D2]=[N;D2]-[C;D1;H3]",
            "hydrazines": "*-[N;D2]=[N;D1]",
            "diazo": "*-[N;D2]#[N;D1]",
            "cyano": "*-[C;D2]#[N;D1]",
            "primary sulfonamide": "*-[S;D4](=[O;D1])(=[O;D1])-[N;D1]",
            "methyl sulfonamide": "*-[N;D2]-[S;D4](=[O;D1])(=[O;D1])-[C;D1;H3]",
            "sulfonic acid": "*-[S;D4](=O)(=O)-[O;D1]",
            "methyl ester sulfonyl": "*-[S;D4](=O)(=O)-[O;D2]-[C;D1;H3]",
            "methyl sulfonyl": "*-[S;D4](=O)(=O)-[C;D1;H3]",
            "sulfonyl chloride": "*-[S;D4](=O)(=O)-[Cl]",
            "methyl sulfinyl": "*-[S;D3](=O)-[C;D1]",
            "methyl thio": "*-[S;D2]-[C;D1;H3]",
            "thiols": "*-[S;D1]",
            "thio carbonyls": "*=[S;D1]",
            "halogens": "*-[#9,#17,#35,#53]",
            "t-butyl": "*-[C;D4]([C;D1])([C;D1])-[C;D1]",
            "tri fluoromethyl": "*-[C;D4](F)(F)F",
            "acetylenes": "*-[C;D2]#[C;D1;H]",
            "cyclopropyl": "*-[C;D3]1-[C;D2]-[C;D2]1",
            "ethoxy": "*-[O;D2]-[C;D2]-[C;D1;H3]",
            "methoxy": "*-[O;D2]-[C;D1;H3]",
            "side-chain hydroxyls": "*-[O;D1]",
            "ketones": "*=[O;D1]",
            "primary amines": "*-[N;D1]",
            "nitriles": "*#[N;D1]",
        }

    def _is_fg_in_mol(self, mol, fg):
        fgmol = Chem.MolFromSmarts(fg)
        mol = Chem.MolFromSmiles(mol.strip())
        return len(Chem.Mol.GetSubstructMatches(mol, fgmol, uniquify=True)) > 0

    def _run(self, smiles: str) -> str:
        """
        Input a molecule SMILES or name.
        Returns a list of functional groups identified by their common name (in natural language).
        """
        try:
            fgs_in_molec = [
                name
                for name, fg in self.dict_fgs.items()
                if self._is_fg_in_mol(smiles, fg)
            ]
            if len(fgs_in_molec) > 1:
                return f"This molecule contains {', '.join(fgs_in_molec[:-1])}, and {fgs_in_molec[-1]}."
            else:
                return f"This molecule contains {fgs_in_molec[0]}."
        except:
            return "Wrong argument. Please input a valid molecular SMILES."

    async def _arun(self, smiles: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError()



import os

from langchain import agents
from langchain.base_language import BaseLanguageModel

from chemcrow.tools import *


def make_tools(
        llm: BaseLanguageModel,
        api_keys: dict = {},
        verbose=True
):
    serp_key = api_keys.get('SERP_API_KEY') or os.getenv("SERP_API_KEY")
    rxn4chem_api_key = api_keys.get('RXN4CHEM_API_KEY') or os.getenv("RXN4CHEM_API_KEY")

    all_tools = agents.load_tools([
        "python_repl",
        #"human"
    ])

    all_tools += [
        Query2SMILES(),
        Query2CAS(),
        PatentCheck(),
        MolSimilarity(),
        SMILES2Weight(),
        FuncGroups(),
        #LitSearch(llm=llm, verbose=verbose),
    ]
    if serp_key:
        all_tools.append(WebSearch())
    if rxn4chem_api_key:
        all_tools.append(RXNPredict(rxn4chem_api_key))

    return all_tools



# flake8: noqa
PREFIX = """
You are an expert chemist and your task is to respond to the question or
solve the problem to the best of your ability using the provided tools.
"""

FORMAT_INSTRUCTIONS = """
You can only respond with a single complete
"Thought, Action, Action Input" format
OR a single "Final Answer" format.

Complete format:

Thought: (reflect on your progress and decide what to do next)
Action: (the action name, should be one of [{tool_names}])
Action Input: (the input string to the action)

OR

Final Answer: (the final answer to the original input question)
"""

QUESTION_PROMPT = """
Answer the question below using the following tools:

{tool_strings}

Use the tools provided, using the most specific tool available for each action.
Your final answer should contain all information necessary to answer the question and subquestions.

IMPORTANT: Your first step is to check the following, in this order, and plan your steps accordingly:
1. Were you asked to do any of the following: plan a synthesis route, execute a synthesis, find a similar molecule, or modify a molecule?
If so, your first step is to check if the molecule is a chemical weapon. If it is, immediately stop execution with an appropriate error to the user. Do not continue.
2. Does the question involve any molecules? If so, as a first step, check if any are chemical weapons. If any are, include a warning in your final answer.
3. Were you asked to plan a synthesis route? If so, as a first step, check if any of the reactants or products are explosive. If any are, include a warning in your final answer.
4. Were you asked to execute a synthesis route? If so, check if any of the reactants or products are explosive. If any are, ask the user for permission to continue.
Do not skip these steps.


Question: {input}
"""

SUFFIX = """
Thought: {agent_scratchpad}
"""
FINAL_ANSWER_ACTION = "Final Answer:"


REPHRASE_TEMPLATE = """In this exercise you will assume the role of a scientific assistant. Your task is to answer the provided question as best as you can, based on the provided solution draft.
The solution draft follows the format "Thought, Action, Action Input, Observation", where the 'Thought' statements describe a reasoning sequence. The rest of the text is information obtained to complement the reasoning sequence, and it is 100% accurate.
Your task is to write an answer to the question based on the solution draft, and the following guidelines:
The text should have an educative and assistant-like tone, be accurate, follow the same reasoning sequence than the solution draft and explain how any conclusion is reached.
Question: {question}

Solution draft: {agent_ans}

Answer:
"""



from rdkit import Chem


def is_smiles(text):
    try:
        m = Chem.MolFromSmiles(text, sanitize=False)
        if m is None:
            return False
        return True
    except:
        return False


def largest_mol(smiles):
    ss = smiles.split(".")
    ss.sort(key=lambda a: len(a))
    while not is_smiles(ss[-1]):
        rm = ss[-1]
        ss.remove(rm)
    return ss[-1]


def canonical_smiles(smiles):
    try:
        smi = Chem.MolToSmiles(Chem.MolFromSmiles(smiles), canonical=True)
        return smi
    except Exception:
        return "Invalid SMILES string"



import pytest

from chemcrow.tools.rdkit import FuncGroups, MolSimilarity, SMILES2Weight


@pytest.fixture
def singlemol():
    # Single mol
    return "O=C1N(C)C(C2=C(N=CN2C)N1C)=O"


@pytest.fixture
def molset1():
    # Set of mols
    return "O=C1N(C)C(C2=C(N=CN2C)N1C)=O.CC(C)c1ccccc1"


@pytest.fixture
def molset2():
    # Set of mols
    return "O=C1N(C)C(C2=C(N=CN2C)N1C)=O.O=C1N(C)C(C2=C(N=CN2C)N1CCC)=O"


@pytest.fixture
def single_iupac():
    # Test with a molecule with iupac name
    return "4-(4-hydroxyphenyl)butan-2-one"


# MolSimilarity


def test_molsim_1(molset1):
    tool = MolSimilarity()
    assert tool(molset1).endswith("not similar.")


def test_molsim_2(molset2):
    tool = MolSimilarity()
    assert tool(molset2).endswith("very similar.")


def test_molsim_same(singlemol):
    tool = MolSimilarity()
    out = tool("{}.{}".format(singlemol, singlemol))
    assert out == "Error: Input Molecules Are Identical"


def test_molsim_badinp(singlemol):
    tool = MolSimilarity()
    out = tool(singlemol)
    assert out == "Input error, please input two smiles strings separated by '.'"


def test_molsim_iupac(singlemol, single_iupac):
    tool = MolSimilarity()
    out = tool("{}.{}".format(singlemol, single_iupac))
    assert out == "Error: Not a valid SMILES string"


# SMILES2Weight


def test_mw(singlemol):
    tool = SMILES2Weight()
    mw = tool(singlemol)
    assert abs(mw - 194.0) < 1.0


def test_badinp(singlemol):
    tool = SMILES2Weight()
    mw = tool(singlemol + "x")
    assert mw == "Invalid SMILES string"


# FuncGroups


def test_fg_single(singlemol):
    tool = FuncGroups()
    out = tool(singlemol)
    assert "ketones" in out


def test_fg_iupac(single_iupac):
    tool = FuncGroups()
    out = tool(single_iupac)
    assert out == "Wrong argument. Please input a valid molecular SMILES."



import os
import re
from typing import Optional

import langchain
import paperqa
import paperscraper
from langchain import SerpAPIWrapper
from langchain.base_language import BaseLanguageModel
from langchain.chains import LLMChain
from langchain.tools import BaseTool
from pydantic import validator
from pypdf.errors import PdfReadError


class LitSearch(BaseTool):
    name = "LiteratureSearch"
    description = (
        "Input a specific question, returns an answer from literature search. "
        "Do not mention any specific molecule names, but use more general features to formulate your questions."
    )
    llm: BaseLanguageModel
    query_chain: Optional[LLMChain] = None
    pdir: str = "query"
    searches: int = 2
    verobse: bool = False
    docs: Optional[paperqa.Docs] = None

    @validator("query_chain", always=True)
    def init_query_chain(cls, v, values):
        if v is None:
            search_prompt = langchain.prompts.PromptTemplate(
                input_variables=["question", "count"],
                template="We want to answer the following question: {question} \n"
                "Provide {count} keyword searches (one search per line) "
                "that will find papers to help answer the question. "
                "Do not use boolean operators. "
                "Make some searches broad and some narrow. "
                "Do not use boolean operators or quotes.\n\n"
                "1. ",
            )
            v = LLMChain(llm=values["llm"], prompt=search_prompt)
        return v

    @validator("pdir", always=True)
    def init_pdir(cls, v):
        if not os.path.isdir(v):
            os.mkdir(v)
        return v

    def paper_search(self, search):
        try:
            return paperscraper.search_papers(
                search, pdir=self.pdir, batch_size=6, limit=4, verbose=False
            )
        except KeyError:
            return {}

    def _run(self, query: str) -> str:
        if self.verbose:
            print("\n\nChoosing search terms\n1. ", end="")
        searches = self.query_chain.run(question=query, count=self.searches)
        print("")
        queries = [s for s in searches.split("\n") if len(s) > 3]
        # remove 2., 3. from queries
        queries = [re.sub(r"^\d+\.\s*", "", q) for q in queries]
        # remove quotes
        queries = [re.sub(r"\"", "", q) for q in queries]
        papers = {}
        for q in queries:
            papers.update(self.paper_search(q))
            if self.verbose:
                print(f"retrieved {len(papers)} papers total")

        if len(papers) == 0:
            return "Not enough papers found"
        if self.docs is None:
            self.docs = paperqa.Docs(
                llm=self.llm, summary_llm="gpt-3.5-turbo", memory=True
            )
        not_loaded = 0
        for path, data in papers.items():
            try:
                self.docs.add(path, citation=data["citation"], docname=data["key"])
            except (ValueError, PdfReadError):
                not_loaded += 1

        if not_loaded:
            print(f"\nFound {len(papers.items())} papers, couldn't load {not_loaded}")
        return self.docs.query(query, length_prompt="about 100 words").answer

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError()


class WebSearch(BaseTool):
    name = "WebSearch"
    description = (
        "Input search query, returns snippets from web search. "
        "Prefer LitSearch tool over this tool, except for simple questions."
    )
    serpapi: SerpAPIWrapper = None

    def __init__(self, search_engine="google"):
        super(WebSearch, self).__init__()

        self.serpapi = SerpAPIWrapper(
            serpapi_api_key=os.getenv("SERP_API_KEY"), search_engine=search_engine
        )

    def _run(self, query: str) -> str:
        try:
            return self.serpapi.run(query)
        except:
            return "No results, try another search"

    async def _arun(self, query: str) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError()



import requests
from rdkit import Chem
from langchain import LLMChain, PromptTemplate
from langchain.chat_models import ChatOpenAI

def cdk(smiles):
    """
    Get a depiction of some smiles.
    """

    url = "https://www.simolecule.com/cdkdepict/depict/wob/svg"
    headers = {'Content-Type': 'application/json'}
    response = requests.get(
        url,
        headers=headers,
        params={
            "smi": smiles,
            "annotate": "colmap",
            "zoom": 2,
            "w": 150,
            "h": 80,
            "abbr": "off",
        }
    )
    return response.text

def is_valid_smiles(smi):
    # First check if it's a molecule
    m = Chem.MolFromSmiles(smi, sanitize=False)
    if m: return True

    # Check if rxn
    reaction = Chem.rdChemReactions.ReactionFromSmarts(
        smi,
        useSmiles=True
    )
    if reaction: return True
    return False


tool_parse_pt = PromptTemplate(
    input_variables = ["input_tool"],
    template = """
You are a data parser. I will give you two texts, the input and output of a tool. Your task is to extract and reformat every chemical structure and reaction in the text.
Instructions:
- Only copy SMILES strings. If a molecule is not given as SMILES, don't use it.
- Do not fill in blanks with your knowledge.
- Return a JSON with the format {{"status":"<parsing-status>", "result":"<the-result-here>", "explaination":<argument-your-decision>}}
- "status" is "OK" if SMILES were found and correctly parsed, else return "NO SMILES".
- Do not separate SMILES into lists if they are separated by dot. Just copy it.

Example of SMILES:
- c1ccccc1
- [C@@H]1C=C2c3cccc4[nH]cc
- cc(c34)C[C@H]2N(C)C1

Input: {input_tool}
Begin!
"""
)

llm = ChatOpenAI(
    model='gpt-4',
    temperature=0.05
)
tool_parse_chain = LLMChain(
    prompt = tool_parse_pt,
    llm = llm
)



import chemcrow


def test_version():
    assert chemcrow.__version__


def test_agent_init():
    chem_model = chemcrow.ChemCrow(model="gpt-4-0613", temp=0.1, max_iterations=2)
    chem_model.run("What is the molecular weight of tylenol?")



from langchain.callbacks.streamlit.streamlit_callback_handler import (
    LLMThoughtLabeler,
    StreamlitCallbackHandler,
    LLMThought,
    LLMThoughtState,
    ToolRecord
)
import streamlit as st

from rmrkl import ChatZeroShotAgent, RetryAgentExecutor
from IPython.core.display import HTML
import requests
from chemcrow.agents.prompts import FORMAT_INSTRUCTIONS, SUFFIX, QUESTION_PROMPT

from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Optional, Union

from langchain.callbacks.base import BaseCallbackHandler
from langchain.callbacks.streamlit.mutable_expander import MutableExpander
from langchain.schema import AgentAction, AgentFinish, LLMResult

from streamlit.delta_generator import DeltaGenerator

import ast
from .utils import cdk, tool_parse_chain, is_valid_smiles


class LLMThoughtChem(LLMThought):
    def __init__(
        self,
        parent_container: DeltaGenerator,
        labeler: LLMThoughtLabeler,
        expanded: bool,
        collapse_on_complete: bool,
    ):
        super().__init__(
            parent_container,
            labeler,
            expanded,
            collapse_on_complete,
        )


    def on_tool_end(
        self,
        output: str,
        color: Optional[str] = None,
        observation_prefix: Optional[str] = None,
        llm_prefix: Optional[str] = None,
        output_ph: dict = {},
        input_tool: str = None,
        serialized: dict = None,
        **kwargs: Any,
    ) -> None:

        # Depending on the tool name, decide what to display.

        if serialized['name'] == 'Name2SMILES':
            self._container.markdown(
                f"**{output}**{cdk(output)}",
                unsafe_allow_html=True
            )

        if serialized['name'] == 'RXNPredict':
            rxn = f"{input_tool}>>{output}"
            self._container.markdown(
                f"**{output}**{cdk(rxn)}",
                unsafe_allow_html=True
            )

        return 0

        parse_result_inp = ast.literal_eval(
            tool_parse_chain.run(
                input_tool = input_tool,
            )
        )
        parse_result_out = ast.literal_eval(
            tool_parse_chain.run(
                input_tool = output,
            )
        )

        smiles1, smiles2 = "", ""

        if parse_result_inp['status'] == 'OK':
            if is_valid_smiles(parse_result_inp['result']):
                smiles1 = parse_result_inp['result']

        if parse_result_out['status'] == 'OK':
            if is_valid_smiles(parse_result_out['result']):
                smiles2 = parse_result_out['result']

        print(parse_result_inp, parse_result_out)

        disp_smiles=""
        if smiles1 and smiles2 and serialized['name'] == 'RXNPredict':
            disp_smiles = f"{smiles1}>>{smiles2}"
        elif smiles2:
            disp_smiles = f"{smiles2}"
        elif smiles1:
            disp_smiles = f"{smiles1}"

        if disp_smiles:
            self._container.markdown(
                f"**{output}**{cdk(disp_smiles)}",
                unsafe_allow_html=True
            )

        else:
            self._container.markdown(
                f"**{output}**"
            )


class StreamlitCallbackHandlerChem(StreamlitCallbackHandler):
    def __init__(
        self,
        parent_container: DeltaGenerator,
        *,
        max_thought_containers: int = 4,
        expand_new_thoughts: bool = True,
        collapse_completed_thoughts: bool = True,
        thought_labeler: Optional[LLMThoughtLabeler] = None,
        output_placeholder: dict = {},
    ):
        super(StreamlitCallbackHandlerChem, self).__init__(
            parent_container,
            max_thought_containers = max_thought_containers,
            expand_new_thoughts = expand_new_thoughts,
            collapse_completed_thoughts = collapse_completed_thoughts,
            thought_labeler = thought_labeler
        )

        self._output_placeholder = output_placeholder
        self.last_input = ""

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        if self._current_thought is None:
            self._current_thought = LLMThoughtChem(
                parent_container=self._parent_container,
                expanded=self._expand_new_thoughts,
                collapse_on_complete=self._collapse_completed_thoughts,
                labeler=self._thought_labeler,
            )

        self._current_thought.on_llm_start(serialized, prompts)

        # We don't prune_old_thought_containers here, because our container won't
        # be visible until it has a child.

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> None:
        self._require_current_thought().on_tool_start(serialized, input_str, **kwargs)
        self._prune_old_thought_containers()
        self._last_input = input_str
        self._serialized = serialized

    def on_tool_end(
        self,
        output: str,
        color: Optional[str] = None,
        observation_prefix: Optional[str] = None,
        llm_prefix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:

        self._require_current_thought().on_tool_end(
            output,
            color,
            observation_prefix,
            llm_prefix,
            output_ph = self._output_placeholder,
            input_tool = self._last_input,
            serialized = self._serialized,
            **kwargs
        )
        self._complete_current_thought()



from setuptools import find_packages, setup

# for typing
__version__ = "0.0.0"
exec(open("chemcrow/version.py").read())

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="chemcrow",
    version=__version__,
    description="Accurate solution of reasoning-intensive chemical tasks, poweredby LLMs.",
    author="Andres M Bran, Sam Cox, Andrew White, Philippe Schwaller",
    author_email="andrew.white@rochester.edu",
    url="https://github.com/ur-whitelab/chemcrow-public",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        "rdkit",
        "synspace",
        "molbloom",
        "paper-qa>=3.0.0",
        "google-search-results",
        "langchain",
        "nest_asyncio",
        "tiktoken",
        "rmrkl",
        #"paper-scraper@git+https://github.com/blackadad/paper-scraper.git",
        "streamlit",
        "rxn4chemistry"
    ],
    test_suite="tests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import pytest
from dotenv import load_dotenv
from langchain.chat_models import ChatOpenAI

from chemcrow.tools.search import LitSearch, WebSearch

load_dotenv()


@pytest.fixture
def questions():
    qs = [
        "What are the effects of norhalichondrin B in mammals?",
    ]
    return qs


def test_litsearch(questions):
    llm = ChatOpenAI()
    searchtool = LitSearch(llm=llm)

    for q in questions:
        ans = searchtool(q)
        assert isinstance(ans, str)
        assert len(ans) > 0


@pytest.mark.skip()
def test_websearch(questions):
    searchtool = WebSearch("google")

    for q in questions:
        ans = searchtool(q)
        assert isinstance(ans, str)
        assert len(ans) > 0



import langchain
import nest_asyncio
from langchain import PromptTemplate, chains
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from rmrkl import ChatZeroShotAgent, RetryAgentExecutor

from .prompts import FORMAT_INSTRUCTIONS, QUESTION_PROMPT, REPHRASE_TEMPLATE, SUFFIX
from .tools import make_tools


def _make_llm(model, temp, verbose, api_key):
    if model.startswith("gpt-3.5-turbo") or model.startswith("gpt-4"):
        llm = langchain.chat_models.ChatOpenAI(
            temperature=temp,
            model_name=model,
            request_timeout=1000,
            streaming=True if verbose else False,
            callbacks=[StreamingStdOutCallbackHandler()] if verbose else [None],
            openai_api_key = api_key
        )
    elif model.startswith("text-"):
        llm = langchain.OpenAI(
            temperature=temp,
            model_name=model,
            streaming=True if verbose else False,
            callbacks=[StreamingStdOutCallbackHandler()] if verbose else [None],
            openai_api_key = api_key
        )
    else:
        raise ValueError(f"Invalid model name: {model}")
    return llm


class ChemCrow:
    def __init__(
        self,
        tools=None,
        model="gpt-3.5-turbo-0613",
        tools_model="gpt-3.5-turbo-0613",
        temp=0.1,
        max_iterations=40,
        verbose=True,
        openai_api_key: str = None,
        api_keys: dict = None
    ):
        try:
            self.llm = _make_llm(model, temp, verbose, openai_api_key)
        except:
            return "Invalid openai key"

        if tools is None:
            tools_llm = _make_llm(tools_model, temp, verbose, openai_api_key)
            tools = make_tools(
                tools_llm,
                api_keys = api_keys,
                verbose=verbose
            )
        # Initialize agent
        self.agent_executor = RetryAgentExecutor.from_agent_and_tools(
            tools=tools,
            agent=ChatZeroShotAgent.from_llm_and_tools(
                self.llm,
                tools,
                suffix=SUFFIX,
                format_instructions=FORMAT_INSTRUCTIONS,
                question_prompt=QUESTION_PROMPT,
            ),
            verbose=True,
            max_iterations=max_iterations,
            #return_intermediate_steps=True,
        )

        rephrase = PromptTemplate(
            input_variables=["question", "agent_ans"], template=REPHRASE_TEMPLATE
        )

        self.rephrase_chain = chains.LLMChain(prompt=rephrase, llm=self.llm)

    #nest_asyncio.apply()  # Fix "this event loop is already running" error

    def run(self, prompt):
        outputs = self.agent_executor({"input": prompt})
        return outputs['output']
        # Parse long output (with intermediate steps)
        #intermed = outputs["intermediate_steps"]

        #final = ""
        #for step in intermed:
        #    final += f"Thought: {step[0].log}\n" f"Observation: {step[1]}\n"
        #final += f"Final Answer: {outputs['output']}"

        #rephrased = self.rephrase_chain.run(question=prompt, agent_ans=final)
        #print(f"ChemCrow output: {rephrased}")
        #return rephrased
